#include "BLEDevice.h"
#include "oled.h"

static BLEUUID bmeServiceUUID("ecc34a9c-e4f4-11ec-8fea-0242ac120002");
static BLEUUID temperature_UUID("ecc34ca4-e4f4-11ec-8fea-0242ac120002");
static BLEUUID pressure_UUID("ecc34dda-e4f4-11ec-8fea-0242ac120002");
static BLEUUID humidity_UUID("ecc34ef2-e4f4-11ec-8fea-0242ac120002");


//Flags stating if should begin connecting and if the connection is up
static boolean doConnect = false;
static boolean connected = false;

//Address of the peripheral device. Address will be found during scanning...
static BLEAddress* pServerAddress = NULL;

//Characteristicd that we want to read
BLERemoteCharacteristic* pTempCharacteristic = NULL;
BLERemoteCharacteristic* pPressCharacteristic = NULL;
BLERemoteCharacteristic* pHumidityCharacteristic = NULL;

//Activate notify
const uint8_t notificationOn[] = {0x1, 0x0};
const uint8_t notificationOff[] = {0x0, 0x0};

//Declaration for an OLED display connected to I2C (SDA, SCL pins)
//OLED display = OLED(21, 22, NO_RESET_PIN, 0x3C, 128, 64, false);   // SSD1306
OLED display = OLED(21, 22, NO_RESET_PIN, 0x3C, 128, 64, true); // SH1106

//Variables to store temperature and humidity
char temperatureChar[6];
char pressureChar[6];
char humidityChar[6];

//Flags to check whether new temperature and humidity readings are available
boolean newTemperature = false;
boolean newPressure = false;
boolean newHumidity = false;

//Connect to the BLE Server that has the name, Service, and Characteristics
bool connectToServer(BLEAddress pAddress) {
  BLEClient* pClient = BLEDevice::createClient();
  // Connect to the remove BLE Server.
  pClient->connect(pAddress);
  Serial.println(" - Connected to server");
  // Obtain a reference to the service we are after in the remote BLE server.
  BLERemoteService* pRemoteService = pClient->getService(bmeServiceUUID);
  if (pRemoteService == nullptr) {
    Serial.print("Failed to find our service UUID: ");
    Serial.println(bmeServiceUUID.toString().c_str());
    return (false);
  }
  // Obtain a reference to the characteristics in the service of the remote BLE server.
  pTempCharacteristic = pRemoteService->getCharacteristic(temperature_UUID);
  pPressCharacteristic = pRemoteService->getCharacteristic(pressure_UUID);
  pHumidityCharacteristic = pRemoteService->getCharacteristic(humidity_UUID);
  if (pTempCharacteristic == nullptr || pHumidityCharacteristic == nullptr || pPressCharacteristic == nullptr) {
    Serial.print("Failed to find our characteristic UUID");
    return false;
  }
  Serial.println(" - Found our characteristics");
  //Assign callback functions for the Characteristics
  pTempCharacteristic->registerForNotify(temperatureNotifyCallback);
  pPressCharacteristic->registerForNotify(pressureNotifyCallback);
  pHumidityCharacteristic->registerForNotify(humidityNotifyCallback);
  return true;
}

//Callback function that gets called, when another device's advertisement has been received
class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      if (advertisedDevice.getName() == "ESP32_BME280") { //Check if the name of the advertiser matches
        advertisedDevice.getScan()->stop(); //Scan can be stopped, we found what we are looking for
        pServerAddress = new BLEAddress(advertisedDevice.getAddress()); //Address of advertiser is the one we need
        doConnect = true; //Set indicator, stating that we are ready to connect
        Serial.println("Device found. Connecting!");
      }
    }
};

//When the BLE Server sends a new temperature reading with the notify property
static void temperatureNotifyCallback(BLERemoteCharacteristic* pBLERemoteCharacteristic,
                                      uint8_t* pData, size_t length, bool isNotify) {
  //store temperature value
  memcpy(temperatureChar,(char*)pData,length);
  newTemperature = true;
}

//When the BLE Server sends a new pressure reading with the notify property
static void pressureNotifyCallback(BLERemoteCharacteristic* pBLERemoteCharacteristic,
                                   uint8_t* pData, size_t length, bool isNotify) {
  //store temperature value
  memcpy(pressureChar,(char*)pData,length);
  newPressure = true;
}
//When the BLE Server sends a new humidity reading with the notify property
static void humidityNotifyCallback(BLERemoteCharacteristic* pBLERemoteCharacteristic,
                                   uint8_t* pData, size_t length, bool isNotify) {
  //store humidity value
  memcpy(humidityChar,(char*)pData,length);
  newHumidity = true;
}

//function that prints the latest sensor readings in the OLED display
void printReadings() {
  char msg[20];
  display.clear();
  Serial.printf("Temperature: %s °C", temperatureChar);
  sprintf(msg, "T:%s%cC", temperatureChar, 133);
  display.draw_string(0, 4, msg, OLED::DOUBLE_SIZE);
  Serial.printf(" - Pressure: %s hPa", pressureChar);
  sprintf(msg, "P:%s hPa", pressureChar);
  display.draw_string(0, 24, msg, OLED::DOUBLE_SIZE);
  Serial.printf(" - Humidity: %s %%\r\n", msg);
  sprintf(msg, "H:%s %%", humidityChar);
  display.draw_string(0, 44, msg, OLED::DOUBLE_SIZE);
  display.display();
}

void setup() {
  //OLED display setup
  display.begin();
  display.clear();
  display.draw_string(4, 12, "BLE Client", OLED::DOUBLE_SIZE);
  display.display();
  //Start serial communication
  Serial.begin(115200);
  Serial.println("Starting Arduino BLE Client application...");

  //Init BLE device
  BLEDevice::init("");

  // Retrieve a Scanner and set the callback we want to use to be informed when we
  // have detected a new device.  Specify that we want active scanning and start the
  // scan to run for 30 seconds.
  BLEScan* pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setActiveScan(true);
  pBLEScan->start(30);
}

void loop() {
  // If the flag "doConnect" is true then we have scanned for and found the desired
  // BLE Server with which we wish to connect.  Now we connect to it.  Once we are
  // connected we set the connected flag to be true.
  if (doConnect == true) {
    if (connectToServer(*pServerAddress)) {
      Serial.println("We are now connected to the BLE Server.");
      //Activate the Notify property of each Characteristic
      pTempCharacteristic->getDescriptor(BLEUUID((uint16_t)0x2902))->writeValue((uint8_t*)notificationOn, 2, true);
      pPressCharacteristic->getDescriptor(BLEUUID((uint16_t)0x2902))->writeValue((uint8_t*)notificationOn, 2, true);
      pHumidityCharacteristic->getDescriptor(BLEUUID((uint16_t)0x2902))->writeValue((uint8_t*)notificationOn, 2, true);
      connected = true;
    } else {
      Serial.println("We have failed to connect to the server; Restart your device to scan for nearby BLE server again.");
    }
    doConnect = false;
  }
  //if new temperature readings are available, print in the OLED
  if (newTemperature && newPressure && newHumidity) {
    newTemperature = false;
    newPressure = false;
    newHumidity = false;
    printReadings();
  }
  delay(1000); // Delay a second between loops.
}
